; FPN_ASM on H144
;
; 29th July 2003
;
; FPN LRESPRs to the function FPNO$(address,[places]).
; This takes one or two parameters, <address>, obligatory
; and <places>, optional.
;
; <address> points to an extended foramt (X) fp number and
; <places> gives the number of places to be printed of the
; mantissa (1 to 21). If <places> is not given the number
; of places defaults to 9.
;
; If the address is odd or the FP number not a true one
; a bad parameter is signalled.
;
; If there is no FPU, FPN returns "not implemented".
;
; Version 2.01 uses a method not relying on packed decimal
; format so that it will work on a 68060.
;
	IN	WIN1_SYS_EQUBAS_ASM
	IN	WIN1_ASS_FP2_EQU_ASM
	IN	WIN1_LIB_HED1
SYS_FPU	EQU	$D0
SYS_CLFP	EQU	$D8
VERSION	EQU	?"2.01"

;
; Check that FPU is present and save/restore is in operation
;
	MOVEQ	#MT_INF,D0
	TRAP	#1
	MOVE.B	SYS_FPU(A0),D0	Type FPU byte
	ANDI.B	#$18,D0
	BEQ	UNIMP		Don't know
	CMPI.B	#$18,D0
	BEQ	UNIMP		no FPU
;
; Link in FPN$
;
	LEA	DEFINE,A1
	MOVEA.W	BP_INIT,A2
	JSR	(A2)

	LEA	MES1,A1
OUT	MOVEA.L	BV_CHBAS(A6),A0
	MOVEA.L	(A0,A6.L),A0
	MOVEA.W	UT_MTEXT,A2
	JSR	(A2)
	MOVEQ	#0,D0
	RTS


;
DEFINE	DC.W	0,0		no procs
	DC.W	1		1 function
	DC.W	FPP-*
	DC.B	5,"FPNO$"
	DC.W	0		mark end
;
VECAD	DC.L	0		place for vector address
;
; Bring in the first parameter (address)
;
FPP	CMPA.L	A3,A5
	BEQ	BADPAR		---->
	MOVEA.L	A5,A4
	LEA	8(A3),A5		arrange to get 1 parameter
	MOVE.L	A5,D5		keep start of next parameter
	MOVEA.W	CA_GTLIN,A2
	JSR	(A2)
	BNE	BADPAR		---->
	MOVEA.L	A4,A5		replace old A5
	MOVE.L	(A6,A1.L),D0	address to D0
	ADDQ.L	#4,BV_RIP(A6)	tidy arith stack
	BTST	#0,D0		even? . . .
	BNE	BADPAR		. . . no
	MOVEA.L	D0,A4		OK address
;
; Now get <places>
;
	MOVEQ	#8,D7		default value
	CMP.L	A5,D5		is there a parameter? . .
	BEQ	FPP_1		. . no
;
	MOVEA.L	D5,A3		pointer to <places>
	LEA	8(A3),A2
	CMPA.L	A2,A5
	BNE	BADPAR		---->
	MOVEA.W	CA_GTINT,A2
	JSR	(A2)
	BNE	BADPAR		---->
	MOVEQ	#0,D7
	MOVE.W	(A6,A1.L),D7	set <places> in D7
	ADDQ.L	#2,BV_RIP(A6)	tidy stack
	SUBQ.W	#1,D7
	BLT	BADPAR		---->
	CMPI.W	#20,D7
	BGT	BADPAR		---->
;
; Now check it is a proper number
;
FPP_1	TST.B	(A4)
	SMI	D3		set sign in D3
	BFEXTS	(A4){1:15},D0       exponent
	BNE	FPP9		not zero
	TST.L	4(A4)
	BNE	FPP9		not a zero mantissa
	TST.L	8(A4)
	BEQ	FPP10		zero found
FPP9	ADDQ.L	#1,D0
	BEQ	BADPAR	---->	either NAN or infinity
	TST.B	4(A4)
	BPL	BADPAR	---->	not normalised
	MOVEQ	#21,D0
	ADD.W	D7,D0		places required - 1
	ADD.L	(A6),D0
	SUB.L	8(A6),D0		shortfall
	BPL	BUFFUL		---->
;
; Set FP number to BASIC buffer in packed decimal form
;
	MOVE.L	VECAD,D6
	BNE	FPP7		already set
	MOVEQ	#MT_INF,D0
	TRAP	#1		A0 = sys_vars
	MOVE.L	SYS_CLFP(A0),D0
	BEQ	BADPAR		---->
	SUBA.L	A0,A0		no user exception routines
	JSR	(D0.L)
	TST.L	D0		OK? . .
	BNE	BADPAR		. . no
	LEA	VECAD,A0
	MOVE.L	A1,(A0)
	MOVE.L	A1,D6
;
; D6 now contains the address of FPSP
;
FPP7	FMOVE.X	(A4),FP7		set n to FP7
	LEA	([A6],A6.L),A4
	FMOVE.X	FP7,(A4)
	FMOVE.L	FPCR,D4
	MOVE.L	D4,D0
	MOVE.B	#$20,D0		rounding down
	FMOVE.L	D0,FPCR
	TST.B	D3
	BEQ	FPP8		positive
	FNEG.X	FP7
FPP8	FMOVE.X	FP7,-(A7)
	JSR	FLOG10X(D6.L)	FP0 contains log10(n)
	FMOVE.X	FP0,(A7)
	JSR	FINTX(D6.L)
	FMOVE.L	FP0,D5		K to D5
	FMOVE.X	FP0,(A7)
	JSR	FTENTOXX(D6.L)
	LEA	12(A7),A7
	FDIV.X	FP0,FP7		r is now in FP7
;
; D5  = K
; FP7 = r
; D7  = p (places required -1)
; (n = 10^K * r)
;
; Now we want to print (sign)r[Ez]
; n = (sign)10^z * r
; Ez is not printed if z=0
; p places are printed (1 to	20)
;


	LEA	([A6],A6.L,12),A2	we are in a Q60 so forget A6 stuff
	TST.B	D3
	BEQ	FPP11
	MOVE.B	#'-',(A2)+
FPP11	FMOVE.B	#10,FP3
	FMOVE.X	FP7,-(A7) 	1st character
	JSR	FINTX(D6.L)
	LEA	12(A7),A7
	FMOVE.B	FP0,D0
	ADDI.B	#'0',D0
	MOVE.B	D0,(A2)+
	FSUB.X	FP0,FP7
	FBEQ	FPP12		ended
	SUBQ.W	#1,D7
	BMI	FPP12		ended
	MOVE.B	#'.',(A2)+
FPP13	FMUL.X	FP3,FP7		get next digit
	FMOVE.X	FP7,-(A7)
	JSR	FINTX(D6.L)
	LEA	12(A7),A7
	FMOVE.B	FP0,D0
	ADDI.B	#'0',D0
	MOVE.B	D0,(A2)+
	FSUB.X	FP0,FP7
	FBEQ	FPP12
	DBF	D7,FPP13
;
; Now print the exponent (if not zero)
;
FPP12	FMOVE.L	D4,FPCR
	TST.W	D5
	BEQ	FPP14
	MOVE.B	#'E',(A2)+
	MOVE.L	D5,D0
	BSR	LTODZ		sets exponent (signed)
FPP14	LEA	([A6],A6.L,12),A3	start of number
	SUBA.L	A3,A2		length of string
	MOVE.L	A2,D4
FPP15	MOVE.L	D4,D1
	ADDQ.L	#3,D1		allow space for length and ..
	BCLR	#0,D1		.. round up to even
	MOVE.L	D1,D5
	MOVEA.W	BV_CHRIX,A1
	JSR	(A1)
	MOVEA.L	BV_RIP(A6),A1	new value of maths stack pointer
	SUBA.L	D5,A1		set to start of FP string
	MOVE.L	A1,BV_RIP(A6)	set it in BASIC VARS
	MOVE.W	D4,(A6,A1.L)	set count of string
	LSR.W	#1,D5		count in words
	MOVEA.L	(A6),A2		point to start of string . .
	LEA	12(A2),A2 	. . rel to A6
	BRA	END
END_1	MOVE.W	(A6,A2.L),(A6,A1.L)
	ADDQ.L	#2,A2
END	ADDQ.L	#2,A1
	DBF	D5,END_1
;
	MOVEQ	#1,D4		mark 'string'
	MOVEQ	#0,D0		good exit
	RTS			back to BASIC
;
; Number is zero
;
FPP10	MOVEQ	#1,D4		length 1
	MOVE.B	#'0',([A6],A6.L,12) set zero as string
	BRA	FPP15




;
BADPAR	MOVEQ	#-15,D0
	RTS
BUFFUL	MOVEQ	#-5,D0
	RTS
UNIMP	LEA	MES2,A1
	BRA	OUT

	HED1	<"George Gwilt's FPNO v",VERSION,10>,MES1
	HED1	<"No FPU found",10>,MES2
;
; LTOD and LTODZ are here
;
	IN	WIN1_LIB_LTOD_SYM_LST
	LIB	WIN1_LIB_LTOD_BIN
